/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#ifndef AIDGE_CPU_OPERATOR_SLICEIMPL_H_
#define AIDGE_CPU_OPERATOR_SLICEIMPL_H_

#include <memory>
#include <tuple>
#include <vector>

#include "aidge/backend/OperatorImpl.hpp"
#include "aidge/operator/Slice.hpp"

#include "aidge/utils/Registrar.hpp"
#include "aidge/utils/Types.h"

namespace Aidge {
// class Slice_Op;

// compute kernel registry for forward and backward
template <DimIdx_t DIM>
class SliceImplForward_cpu
    : public Registrable<SliceImplForward_cpu<DIM>, std::tuple<DataType>,
                         void(const typename Slice_Op<DIM>::Attrs&,
                              const std::array<std::size_t, DIM>,
                              const void*,
                              void*)> {};
template <DimIdx_t DIM>
class SliceImplBackward_cpu
    : public Registrable<SliceImplBackward_cpu<DIM>, std::tuple<DataType>,
                         void(const typename Slice_Op<DIM>::Attrs&,
                              const std::array<std::size_t, DIM>,
                              const void*,
                              void*)> {};

template <DimIdx_t DIM>
class SliceImpl_cpu : public OperatorImpl {
   private:
    const Slice_Op<DIM>& mOp;
    std::array<NbElts_t, 1> mNbConsumedData;
    std::array<NbElts_t, 1> mNbProducedData;

   public:
    SliceImpl_cpu(const Slice_Op<DIM>& op) : mOp(op), mNbConsumedData({0}), mNbProducedData({0}) {}

    static std::unique_ptr<SliceImpl_cpu<DIM>> create(const Slice_Op<DIM>& op) {
        return std::make_unique<SliceImpl_cpu<DIM>>(op);
    }

   public:
    NbElts_t getNbRequiredData(const IOIndex_t /*inputIdx*/) const override final {
        assert(mOp.getInput(0) && "requires valid input");

        // Requires the whole tensors
        const auto& inputDims = mOp.getInput(0)->dims();

        return std::accumulate(inputDims.begin(), inputDims.end(), static_cast<NbElts_t>(1),
                               std::multiplies<NbElts_t>());
    }
    NbElts_t getNbRequiredProtected(const IOIndex_t /*inputIdx*/) const override final { return 0; }
    NbElts_t getRequiredMemory(const IOIndex_t outputIdx,
                               const std::vector<DimSize_t>& inputsSize) const override final {
        (void)outputIdx;
        (void)inputsSize;
        const auto& outputDims = mOp.getOutput(0)->dims();
        return std::accumulate(outputDims.begin(), outputDims.end(), static_cast<NbElts_t>(1),
                               std::multiplies<NbElts_t>());
    }
    NbElts_t getNbConsumedData(const IOIndex_t /*inputIdx*/) const override final {
        return mNbConsumedData[0];
    }
    NbElts_t getNbProducedData(const IOIndex_t outputIdx) const override final {
        return mNbProducedData[0];
    }
    void updateConsummerProducer() override final {
        // each input is consumed by the minimum amount for a forward pass
        mNbConsumedData[0] += getNbRequiredData(0);

        mNbProducedData[0] += getRequiredMemory(0, {});
    }

    void forward() {
        // FIXME: uncomment the following code once memory handling will work
        assert(mOp.getInput(0) && "missing input #0");

        // Find the correct kernel type
        auto kernelFunc = Registrar<SliceImplForward_cpu<DIM>>::create(
                {mOp.getInput(0)->dataType()});

        // Call kernel
        kernelFunc(mOp.getInput(0)->template dims<DIM>(),
                   std::get<1>(mOp.getStaticAttributes()),
                   mOp.getInput(0)->getImpl()->rawPtr(),
                   mOp.getOutput(0)->getImpl()->rawPtr()
                );

        // each input is consumed by the minimum amount for a forward pass
        mNbConsumedData[0] += getNbRequiredData(0);

        mNbProducedData[0] += getRequiredMemory(0, {});
    }

    void backward() { printf("Not implemented yet.\n"); }
};

/******************************************************************************/

template <>
class SliceImpl_cpu<1> : public OperatorImpl {
   private:
    const Slice_Op<1>& mOp;
    std::array<NbElts_t, 1> mNbConsumedData;
    std::array<NbElts_t, 1> mNbProducedData;

   public:
    SliceImpl_cpu(const Slice_Op<1>& op) : mOp(op), mNbConsumedData({0}), mNbProducedData({0}) {}

    static std::unique_ptr<SliceImpl_cpu<1>> create(const Slice_Op<1>& op) {
        return std::make_unique<SliceImpl_cpu<1>>(op);
    }

   public:
    NbElts_t getNbRequiredData(const IOIndex_t /*inputIdx*/) const override final;
    NbElts_t getNbRequiredProtected(const IOIndex_t /*inputIdx*/) const override final;
    NbElts_t getRequiredMemory(const IOIndex_t outputIdx,
                               const std::vector<DimSize_t>& inputsSize) const override final;
    NbElts_t getNbConsumedData(const IOIndex_t /*inputIdx*/) const override final;
    NbElts_t getNbProducedData(const IOIndex_t outputIdx) const override final;
    void updateConsummerProducer() override final;

    void forward();
    void backward();
};

/******************************************************************************/

template <>
class SliceImpl_cpu<2> : public OperatorImpl {
   private:
    const Slice_Op<2>& mOp;
    std::array<NbElts_t, 1> mNbConsumedData;
    std::array<NbElts_t, 1> mNbProducedData;

   public:
    SliceImpl_cpu(const Slice_Op<2>& op) : mOp(op), mNbConsumedData({0}), mNbProducedData({0}) {}

    static std::unique_ptr<SliceImpl_cpu<2>> create(const Slice_Op<2>& op) {
        return std::make_unique<SliceImpl_cpu<2>>(op);
    }

   public:
    NbElts_t getNbRequiredData(const IOIndex_t /*inputIdx*/) const override final;
    NbElts_t getNbRequiredProtected(const IOIndex_t /*inputIdx*/) const override final;
    NbElts_t getRequiredMemory(const IOIndex_t outputIdx,
                               const std::vector<DimSize_t>& inputsSize) const override final;
    NbElts_t getNbConsumedData(const IOIndex_t /*inputIdx*/) const override final;
    NbElts_t getNbProducedData(const IOIndex_t outputIdx) const override final;
    void updateConsummerProducer() override final;

    void forward();
    void backward();
};

/******************************************************************************/

template <>
class SliceImpl_cpu<3> : public OperatorImpl {
   private:
    const Slice_Op<3>& mOp;
    std::array<NbElts_t, 1> mNbConsumedData;
    std::array<NbElts_t, 1> mNbProducedData;

   public:
    SliceImpl_cpu(const Slice_Op<3>& op) : mOp(op), mNbConsumedData({0}), mNbProducedData({0}) {}

    static std::unique_ptr<SliceImpl_cpu<3>> create(const Slice_Op<3>& op) {
        return std::make_unique<SliceImpl_cpu<3>>(op);
    }

   public:
    NbElts_t getNbRequiredData(const IOIndex_t /*inputIdx*/) const override final;
    NbElts_t getNbRequiredProtected(const IOIndex_t /*inputIdx*/) const override final;
    NbElts_t getRequiredMemory(const IOIndex_t outputIdx,
                               const std::vector<DimSize_t>& inputsSize) const override final;
    NbElts_t getNbConsumedData(const IOIndex_t /*inputIdx*/) const override final;
    NbElts_t getNbProducedData(const IOIndex_t outputIdx) const override final;
    void updateConsummerProducer() override final;

    void forward();
    void backward();
};

/******************************************************************************/

template <>
class SliceImpl_cpu<4> : public OperatorImpl {
   private:
    const Slice_Op<4>& mOp;
    std::array<NbElts_t, 1> mNbConsumedData;
    std::array<NbElts_t, 1> mNbProducedData;

   public:
    SliceImpl_cpu(const Slice_Op<4>& op) : mOp(op), mNbConsumedData({0}), mNbProducedData({0}) {}

    static std::unique_ptr<SliceImpl_cpu<4>> create(const Slice_Op<4>& op) {
        return std::make_unique<SliceImpl_cpu<4>>(op);
    }

   public:
    NbElts_t getNbRequiredData(const IOIndex_t /*inputIdx*/) const override final;
    NbElts_t getNbRequiredProtected(const IOIndex_t /*inputIdx*/) const override final;
    NbElts_t getRequiredMemory(const IOIndex_t outputIdx,
                               const std::vector<DimSize_t>& inputsSize) const override final;
    NbElts_t getNbConsumedData(const IOIndex_t /*inputIdx*/) const override final;
    NbElts_t getNbProducedData(const IOIndex_t outputIdx) const override final;
    void updateConsummerProducer() override final;

    void forward();
    void backward();
};



namespace {
static Registrar<Slice_Op<1>> registrarSliceImpl_1D_cpu("cpu", Aidge::SliceImpl_cpu<1>::create);
static Registrar<Slice_Op<2>> registrarSliceImpl_2D_cpu("cpu", Aidge::SliceImpl_cpu<2>::create);
static Registrar<Slice_Op<3>> registrarSliceImpl_3D_cpu("cpu", Aidge::SliceImpl_cpu<3>::create);
static Registrar<Slice_Op<4>> registrarSliceImpl_4D_cpu("cpu", Aidge::SliceImpl_cpu<4>::create);
}  // namespace
}  // namespace Aidge

#endif /* AIDGE_CPU_OPERATOR_LEAKYRELUIMPL_H_ */
