/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#ifndef AIDGE_CPU_OPERATOR_SLICEIMPL_FORWARD_KERNEL_H_
#define AIDGE_CPU_OPERATOR_SLICEIMPL_FORWARD_KERNEL_H_

#include "aidge/utils/Registrar.hpp"
#include "aidge/operator/Slice.hpp"
#include "aidge/backend/cpu/operator/SliceImpl.hpp"
#include <array>
#include <cstddef>

#include "aidge/data/Data.hpp"

namespace Aidge {
template <class I, std::size_t DIM>
void SliceImpl_cpu_forward_kernel(const typename Slice_Op<DIM>::Attrs& attrs,
                                     const std::array<std::size_t, DIM> inputDims,
                                     const void* input_,
                                     void* output_) {

    const I* input = static_cast<const I*>(input_) + std::get<0>(attrs);
    I* output = static_cast<I*>(output_);
	const std::array<std::size_t, DIM> slicedDims = std::get<1>(attrs);

	// for inputDims = {4,5,5,3} & slicedDims = {3,2,2,1}, substractDims = {1,5,5,3}
    std::array<std::size_t, DIM> substractedDims;
    for (std::size_t i = 0; i < DIM; ++i) {
        substractedDims[i] = inputDims[i] - slicedDims[i];
    }

	// for slicedDims = {3,2,2,1}, prodSlicedDims = {12,4,2,1}
    std::array<std::size_t, DIM> prodSlicedDims;
    std::array<std::size_t, DIM+1> prodInputDims;
	prodSlicedDims[DIM - 1] = slicedDims[DIM - 1];
	prodInputDims[DIM - 1] = inputDims[DIM - 1];
	prodInputDims[DIM] = 1;
	for (std::size_t i = 2; i <= DIM; ++i) {
		prodSlicedDims[DIM - i] = prodSlicedDims[DIM - i + 1]*slicedDims[DIM - i];
		prodInputDims[DIM - i] = prodInputDims[DIM - i + 1]*inputDims[DIM - i];
	}

	std::size_t j = 0;
	std::size_t i = 0;
	for (; j < prodSlicedDims[0];) {
		output[j] = input[i++];
                ++j;
		for (std::size_t idx = DIM - 1; idx > 0; --idx) {
				i += j % prodSlicedDims[idx] == 0 ? substractedDims[idx]*prodInputDims[idx+1] : 0;
		}
	}
}

namespace {

// DIM = 1
static Registrar<SliceImplForward_cpu<1>> registrarSliceImplForward_1D_cpu_Float32(
        {DataType::Float32}, Aidge::SliceImpl_cpu_forward_kernel<float, 1>);
static Registrar<SliceImplForward_cpu<1>> registrarSliceImplForward_1D_cpu_Int32(
        {DataType::Int32}, Aidge::SliceImpl_cpu_forward_kernel<int, 1>);
static Registrar<SliceImplForward_cpu<1>> registrarSliceImplForward_1D_cpu_Float64(
        {DataType::Float64}, Aidge::SliceImpl_cpu_forward_kernel<double, 1>);

// DIM = 2
static Registrar<SliceImplForward_cpu<2>> registrarSliceImplForward_2D_cpu_Float32(
        {DataType::Float32}, Aidge::SliceImpl_cpu_forward_kernel<float, 2>);
static Registrar<SliceImplForward_cpu<2>> registrarSliceImplForward_2D_cpu_Int32(
        {DataType::Int32}, Aidge::SliceImpl_cpu_forward_kernel<int, 2>);
static Registrar<SliceImplForward_cpu<2>> registrarSliceImplForward_2D_cpu_Float64(
        {DataType::Float64}, Aidge::SliceImpl_cpu_forward_kernel<double, 2>);

// DIM = 3
static Registrar<SliceImplForward_cpu<3>> registrarSliceImplForward_3D_cpu_Float32(
        {DataType::Float32}, Aidge::SliceImpl_cpu_forward_kernel<float, 3>);
static Registrar<SliceImplForward_cpu<3>> registrarSliceImplForward_3D_cpu_Int32(
        {DataType::Int32}, Aidge::SliceImpl_cpu_forward_kernel<int, 3>);
static Registrar<SliceImplForward_cpu<3>> registrarSliceImplForward_3D_cpu_Float64(
        {DataType::Float64}, Aidge::SliceImpl_cpu_forward_kernel<double, 3>);

// DIM = 4
static Registrar<SliceImplForward_cpu<4>> registrarSliceImplForward_4D_cpu_Float32(
        {DataType::Float32}, Aidge::SliceImpl_cpu_forward_kernel<float, 4>);
static Registrar<SliceImplForward_cpu<4>> registrarSliceImplForward_4D_cpu_Int32(
        {DataType::Int32}, Aidge::SliceImpl_cpu_forward_kernel<int, 4>);
static Registrar<SliceImplForward_cpu<4>> registrarSliceImplForward_4D_cpu_Float64(
        {DataType::Float64}, Aidge::SliceImpl_cpu_forward_kernel<double, 4>);
}  // namespace
}  // namespace Aidge

#endif /* AIDGE_CPU_OPERATOR_LEAKYRELUIMPL_FORWARD_KERNEL_H_ */
