/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#include <cassert>
#include <numeric> // std::accumulate
#include <vector>

#include "aidge/utils/Types.h"
#include "aidge/data/Data.hpp"
#include "aidge/data/Tensor.hpp"

#include "aidge/backend/cpu/operator/ConcatImpl.hpp"
#include "aidge/backend/cpu/operator/ConcatImpl_forward_kernels.hpp"

Aidge::NbElts_t Aidge::ConcatImpl_cpu::getNbRequiredData(const Aidge::IOIndex_t inputIdx) const {
    assert(mOp.getInput(inputIdx) && "requires valid input");

    // Requires the whole tensors
    const auto& inputDims = std::static_pointer_cast<Tensor>(mOp.getInput(inputIdx))->dims();
    return std::accumulate(inputDims.begin(), inputDims.end(), NbElts_t(1), std::multiplies<NbElts_t>());
}

Aidge::NbElts_t  Aidge::ConcatImpl_cpu::getNbRequiredProtected(const Aidge::IOIndex_t /*inputIdx*/) const {
    // for the direct convolution algorithm, convolutions can be in-place, if there is no padding!
    return 0;
}

Aidge::NbElts_t  Aidge::ConcatImpl_cpu::getRequiredMemory(const Aidge::IOIndex_t outputIdx, const std::vector<Aidge::DimSize_t>& /*inputsSize*/) const {
    // Requires the whole tensors, regardless of available data on inputs
    assert(outputIdx == 0 && "operator has only one output");
    (void) outputIdx;

    const auto& outputDims = std::static_pointer_cast<Tensor>(mOp.getOutput(0))->dims();
    return std::accumulate(outputDims.begin(), outputDims.end(), NbElts_t(1), std::multiplies<NbElts_t>());
}

Aidge::NbElts_t  Aidge::ConcatImpl_cpu::getNbConsumedData(const Aidge::IOIndex_t inputIdx) const {
    assert(inputIdx < mNbConsumedData.size());
    return mNbConsumedData[inputIdx];
}

Aidge::NbElts_t  Aidge::ConcatImpl_cpu::getNbProducedData(const Aidge::IOIndex_t outputIdx) const {
    assert(outputIdx < mNbProducedData.size());
    return mNbProducedData[outputIdx];
}

void  Aidge::ConcatImpl_cpu::updateConsummerProducer() {
    for (IOIndex_t inputIdx = 0; static_cast<NbElts_t>(inputIdx) < mNbConsumedData.size(); ++inputIdx)
        mNbConsumedData[inputIdx]+= getNbRequiredData(inputIdx); // each input is consumed by the minimum amount for a forward pass

    mNbProducedData[0]+= getRequiredMemory(0, {});

}

void  Aidge::ConcatImpl_cpu::forward() {
    assert(mOp.getInput(0) && "missing input in Concat operator");
    DataType datatypeFirstInput = mOp.getInput(0)->dataType();
    for (IOIndex_t i = 1; i < mOp.nbInputs(); ++i) {
        assert(mOp.getInput(i) && "missing input in Concat operator");
        assert(mOp.getInput(i)->dataType() == datatypeFirstInput);
    }

    auto kernelFunc = Registrar<ConcatImplForward_cpu>::create({
        datatypeFirstInput,
        mOp.getOutput(0)->dataType()});

    std::vector<const void*> opInputs;
    for (IOIndex_t i = 0; i < mOp.nbInputs(); ++i) {
        opInputs.push_back(mOp.getInput(i)->getImpl()->rawPtr());
    }

    kernelFunc(mOp.getStaticAttributes(),
               mOp.getInput(0)->dims(),
               opInputs,
               mOp.getOutput(0)->getImpl()->rawPtr());
}

void  Aidge::ConcatImpl_cpu::backward() { printf("Not implemented yet.\n"); }