/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#include <cassert>
#include <numeric>    // std::accumulate
#include <functional> // std::multiplies

#include "aidge/operator/Slice.hpp"

#include "aidge/backend/cpu/operator/SliceImpl.hpp"
#include "aidge/backend/cpu/operator/SliceImpl_forward_kernels.hpp"
#include "aidge/utils/Types.h"
#include <vector>
#include <cassert>
#include <tuple>


Aidge::NbElts_t Aidge::SliceImpl_cpu<1>::getNbRequiredData(const Aidge::IOIndex_t /*inputIdx*/) const {
    assert(mOp.getInput(0) && "requires valid input");

    // Requires the whole tensors
    return mOp.getInput(0)->dims<1>()[0];
}

Aidge::NbElts_t Aidge::SliceImpl_cpu<1>::getNbRequiredProtected(const Aidge::IOIndex_t /*inputIdx*/) const { return 0; }

Aidge::NbElts_t Aidge::SliceImpl_cpu<1>::getRequiredMemory(const Aidge::IOIndex_t outputIdx,
                               const std::vector<Aidge::DimSize_t>& inputsSize) const {
    (void)outputIdx;
    (void)inputsSize;
    return mOp.getOutput(0)->dims<1>()[0];
}

Aidge::NbElts_t Aidge::SliceImpl_cpu<1>::getNbConsumedData(const Aidge::IOIndex_t /*inputIdx*/) const {
    return mNbConsumedData[0];
}

Aidge::NbElts_t Aidge::SliceImpl_cpu<1>::getNbProducedData(const Aidge::IOIndex_t /*outputIdx*/) const {
    return mNbProducedData[0];
}

void Aidge::SliceImpl_cpu<1>::updateConsummerProducer() {
    // each input is consumed by the minimum amount for a forward pass
    mNbConsumedData[0] += getNbRequiredData(0);

    mNbProducedData[0] += getRequiredMemory(0, {});
}

void Aidge::SliceImpl_cpu<1>::forward() {
    // FIXME: uncomment the following code once memory handling will work
    assert(mOp.getInput(0) && "missing input #0");

    // Find the correct kernel type
    auto kernelFunc = Registrar<SliceImplForward_cpu<1>>::create(
            {mOp.getInput(0)->dataType()});

    // Call kernel
    kernelFunc(mOp.getStaticAttributes(),
                mOp.getInput(0)->template dims<1>(),
                mOp.getInput(0)->getImpl()->rawPtr(),
                mOp.getOutput(0)->getImpl()->rawPtr()
            );

    // each input is consumed by the minimum amount for a forward pass
    mNbConsumedData[0] += getNbRequiredData(0);

    mNbProducedData[0] += getRequiredMemory(0, {});
}

void Aidge::SliceImpl_cpu<1>::backward() { printf("Not implemented yet.\n"); }

/////////////////////////////////////////////////////////////////////////

Aidge::NbElts_t Aidge::SliceImpl_cpu<2>::getNbRequiredData(const Aidge::IOIndex_t /*inputIdx*/) const {
    assert(mOp.getInput(0) && "requires valid input");

    // Requires the whole tensors
    const auto& inputDims = mOp.getInput(0)->dims<2>();
    return inputDims[0]*inputDims[1];
}

Aidge::NbElts_t Aidge::SliceImpl_cpu<2>::getNbRequiredProtected(const Aidge::IOIndex_t /*inputIdx*/) const { return 0; }

Aidge::NbElts_t Aidge::SliceImpl_cpu<2>::getRequiredMemory(const Aidge::IOIndex_t outputIdx,
                               const std::vector<Aidge::DimSize_t>& inputsSize) const {
    (void)outputIdx;
    (void)inputsSize;
    const auto& outputDims = mOp.getOutput(0)->dims<2>();
    return outputDims[0]*outputDims[1];
}

Aidge::NbElts_t Aidge::SliceImpl_cpu<2>::getNbConsumedData(const Aidge::IOIndex_t /*inputIdx*/) const {
    return mNbConsumedData[0];
}

Aidge::NbElts_t Aidge::SliceImpl_cpu<2>::getNbProducedData(const Aidge::IOIndex_t /*outputIdx*/) const {
    return mNbProducedData[0];
}

void Aidge::SliceImpl_cpu<2>::updateConsummerProducer() {
    // each input is consumed by the minimum amount for a forward pass
    mNbConsumedData[0] += getNbRequiredData(0);

    mNbProducedData[0] += getRequiredMemory(0, {});
}

void Aidge::SliceImpl_cpu<2>::forward() {
    // FIXME: uncomment the following code once memory handling will work
    assert(mOp.getInput(0) && "missing input #0");

    // Find the correct kernel type
    auto kernelFunc = Registrar<SliceImplForward_cpu<2>>::create(
            {mOp.getInput(0)->dataType()});

    // Call kernel
    kernelFunc(mOp.getStaticAttributes(),
                mOp.getInput(0)->template dims<2>(),
                mOp.getInput(0)->getImpl()->rawPtr(),
                mOp.getOutput(0)->getImpl()->rawPtr()
            );

    // each input is consumed by the minimum amount for a forward pass
    mNbConsumedData[0] += getNbRequiredData(0);

    mNbProducedData[0] += getRequiredMemory(0, {});
}

void Aidge::SliceImpl_cpu<2>::backward() { printf("Not implemented yet.\n"); }

////////////////////////////////////////////////////////////////////////////

Aidge::NbElts_t Aidge::SliceImpl_cpu<3>::getNbRequiredData(const Aidge::IOIndex_t /*inputIdx*/) const {
    assert(mOp.getInput(0) && "requires valid input");

    // Requires the whole tensors
    const auto& inputDims = mOp.getInput(0)->dims<3>();

    return std::accumulate(inputDims.begin(), inputDims.end(), static_cast<NbElts_t>(1),
                            std::multiplies<NbElts_t>());
}

Aidge::NbElts_t Aidge::SliceImpl_cpu<3>::getNbRequiredProtected(const Aidge::IOIndex_t /*inputIdx*/) const { return 0; }

Aidge::NbElts_t Aidge::SliceImpl_cpu<3>::getRequiredMemory(const Aidge::IOIndex_t outputIdx,
                               const std::vector<Aidge::DimSize_t>& inputsSize) const {
    (void)outputIdx;
    (void)inputsSize;
    const auto& outputDims = mOp.getOutput(0)->dims<3>();
    return std::accumulate(outputDims.begin(), outputDims.end(), static_cast<NbElts_t>(1),
                            std::multiplies<NbElts_t>());
}

Aidge::NbElts_t Aidge::SliceImpl_cpu<3>::getNbConsumedData(const Aidge::IOIndex_t /*inputIdx*/) const {
    return mNbConsumedData[0];
}

Aidge::NbElts_t Aidge::SliceImpl_cpu<3>::getNbProducedData(const Aidge::IOIndex_t /*outputIdx*/) const {
    return mNbProducedData[0];
}

void Aidge::SliceImpl_cpu<3>::updateConsummerProducer() {
    // each input is consumed by the minimum amount for a forward pass
    mNbConsumedData[0] += getNbRequiredData(0);

    mNbProducedData[0] += getRequiredMemory(0, {});
}

void Aidge::SliceImpl_cpu<3>::forward() {
    // FIXME: uncomment the following code once memory handling will work
    assert(mOp.getInput(0) && "missing input #0");

    // Find the correct kernel type
    auto kernelFunc = Registrar<SliceImplForward_cpu<3>>::create(
            {mOp.getInput(0)->dataType()});

    // Call kernel
    kernelFunc(mOp.getStaticAttributes(),
                mOp.getInput(0)->template dims<3>(),
                mOp.getInput(0)->getImpl()->rawPtr(),
                mOp.getOutput(0)->getImpl()->rawPtr()
            );

    // each input is consumed by the minimum amount for a forward pass
    mNbConsumedData[0] += getNbRequiredData(0);

    mNbProducedData[0] += getRequiredMemory(0, {});
}

void Aidge::SliceImpl_cpu<3>::backward() { printf("Not implemented yet.\n"); }

//////////////////////////////////////////////////////////////////////////////

Aidge::NbElts_t Aidge::SliceImpl_cpu<4>::getNbRequiredData(const Aidge::IOIndex_t /*inputIdx*/) const {
    assert(mOp.getInput(0) && "requires valid input");

    // Requires the whole tensors
    const auto& inputDims = mOp.getInput(0)->dims<4>();

    return std::accumulate(inputDims.begin(), inputDims.end(), static_cast<NbElts_t>(1),
                            std::multiplies<NbElts_t>());
}

Aidge::NbElts_t Aidge::SliceImpl_cpu<4>::getNbRequiredProtected(const Aidge::IOIndex_t /*inputIdx*/) const { return 0; }

Aidge::NbElts_t Aidge::SliceImpl_cpu<4>::getRequiredMemory(const Aidge::IOIndex_t outputIdx,
                               const std::vector<Aidge::DimSize_t>& inputsSize) const {
    (void)outputIdx;
    (void)inputsSize;
    const auto& outputDims = mOp.getOutput(0)->template dims<4>();
    return std::accumulate(outputDims.begin(), outputDims.end(), static_cast<NbElts_t>(1),
                            std::multiplies<NbElts_t>());
}

Aidge::NbElts_t Aidge::SliceImpl_cpu<4>::getNbConsumedData(const Aidge::IOIndex_t /*inputIdx*/) const {
    return mNbConsumedData[0];
}

Aidge::NbElts_t Aidge::SliceImpl_cpu<4>::getNbProducedData(const Aidge::IOIndex_t /*outputIdx*/) const {
    return mNbProducedData[0];
}

void Aidge::SliceImpl_cpu<4>::updateConsummerProducer() {
    // each input is consumed by the minimum amount for a forward pass
    mNbConsumedData[0] += getNbRequiredData(0);

    mNbProducedData[0] += getRequiredMemory(0, {});
}

void Aidge::SliceImpl_cpu<4>::forward() {
    // FIXME: uncomment the following code once memory handling will work
    assert(mOp.getInput(0) && "missing input #0");

    // Find the correct kernel type
    auto kernelFunc = Registrar<SliceImplForward_cpu<4>>::create(
            {mOp.getInput(0)->dataType()});

    // Call kernel
    kernelFunc(mOp.getStaticAttributes(),
                mOp.getInput(0)->template dims<4>(),
                mOp.getInput(0)->getImpl()->rawPtr(),
                mOp.getOutput(0)->getImpl()->rawPtr()
            );

    // each input is consumed by the minimum amount for a forward pass
    mNbConsumedData[0] += getNbRequiredData(0);

    mNbProducedData[0] += getRequiredMemory(0, {});
}

void Aidge::SliceImpl_cpu<4>::backward() { printf("Not implemented yet.\n"); }